# include "define.inc"
!
! TT: Concerns: In Cray NF90_FLOAT also comes as 8 byte real?
!
! TT: To do: use nf90_inquire_variable to resolve
! TT: the precision mismatch when read
!
module agk_save

# ifdef NETCDF
  use netcdf, only: NF90_FLOAT, NF90_DOUBLE, NF90_NOWRITE, NF90_CLOBBER, NF90_NOERR
  use netcdf, only: nf90_create, nf90_open, nf90_sync, nf90_close
  use netcdf, only: nf90_def_dim, nf90_def_var, nf90_enddef
  use netcdf, only: nf90_put_var, nf90_get_var, nf90_strerror
  use netcdf, only: nf90_inq_dimid, nf90_inquire_dimension
  use netcdf, only: nf90_inq_varid, nf90_inquire_variable

  use netcdf_utils, only: get_netcdf_code_precision
  use netcdf_utils, only: check_netcdf_file_precision
  use netcdf_utils, only: netcdf_error
  use netcdf_utils, only: netcdf_real, netcdf_int, kind_nf
# endif

  implicit none

  public :: agk_restore, agk_save_for_restart
  public :: agk_save_for_restart_energy, agk_restore_energy
  public :: init_save, init_dt, init_tstart, init_ant_amp
!# ifdef NETCDF
!  public :: netcdf_real, kind_nf, get_netcdf_code_precision, netcdf_error
!# endif

  interface agk_restore
     module procedure agk_restore_many!, agk_restore_one
  end interface

  private
  character (300), save :: restart_file
  character (300), save :: restart_file_eq
  character (300), save :: restart_file_ene

  integer, save  :: ndigits           !Number of digits to keep in .nc.* restart filenames
  logical, save  :: separate_em_in    !T=read EM fields from filename.nc.em
  logical, save  :: separate_em_out   !T=write EM fields to filename.nc.em
 
# ifdef NETCDF
  real, allocatable, dimension(:,:,:) :: tmpr, tmpi, ftmpr, ftmpi
  real, allocatable, dimension(:) :: atmp
!  integer, parameter :: kind_nf = kind (NF90_NOERR)
  integer (kind_nf) :: ncid, thetaid, signid, gloid, kyid, kxid, nk_stir_dim
  integer (kind_nf) :: phir_id, phii_id, aparr_id, apari_id, bparr_id, bpari_id
  integer (kind_nf) :: delt0id, t0id, gr_id, gi_id
  integer (kind_nf) :: a_antr_id, b_antr_id, a_anti_id, b_anti_id
  integer (kind_nf) :: nseed_dim, init_seed_id, fin_seed_id

  integer (kind_nf) :: nceqid, nceqgid
  integer (kind_nf) :: thetaid_fm, thetaid_g, riid_fm, riid_g
  integer (kind_nf) :: geq_id
  integer (kind_nf) :: phi_eq_id, apar_eq_id, bpar_eq_id
!  integer (kind_nf) :: netcdf_real=0
  integer (kind_nf) :: ncenid
  integer (kind_nf) :: array_id, arraydim

  logical :: initialized = .false.
  logical :: debug = .false.
# endif

contains

  subroutine agk_save_for_restart &
       (g, t0, delt0, istatus, use_Phi, use_Apar, use_Bpar, exit_in)

# ifdef NETCDF
    use constants, only: kind_rs, kind_rd, pi
    use fields_arrays, only: phinew, aparnew, bparnew
    use kgrids, only: naky, nakx
    use antenna_data, only: nk_stir, a_ant, b_ant, ant_on, nseed, &
         init_seed, fin_seed
    use agk_mem, only: alloc8, dealloc8
# endif
    use mp, only: nproc, iproc, proc0
    use theta_grid, only: ntgrid
! Must include g_layout_type here to avoid obscure bomb while compiling
! agk_diagnostics.f90 (which uses this module) with the Compaq F90 compiler:
    use agk_layouts, only: g_lo
    use layouts_type, only: g_layout_type
    use file_utils, only: error_unit
    implicit none
    complex, dimension (-ntgrid:,:,g_lo%llim_proc:), intent (in) :: g
    real, intent (in) :: t0, delt0
    logical, intent (in) :: use_Phi, use_Apar, use_Bpar
# ifdef NETCDF
    integer, intent (out) :: istatus

# else
    integer :: istatus
# endif
    logical, intent (in), optional :: exit_in
# ifdef NETCDF
    character (306) :: file_proc
    character (10) :: suffix
    character (10)  :: suffix_format
    integer :: i, n_elements, ierr
    logical :: exit
    real, allocatable, dimension(:) :: r_init_seed, r_fin_seed

    if (present(exit_in)) then
       exit = exit_in
    else
       exit = .false.
    end if

    n_elements = g_lo%ulim_proc-g_lo%llim_proc+1
    if (n_elements <= 0) return

    if (.not.initialized) then
       initialized = .true.
       file_proc = trim(restart_file)

       ! RN> This complication of naming scheme is for backward compat.
       call get_suffix_format(nproc,suffix_format)
       write (suffix,suffix_format) '.', iproc
!       write (suffix,'(a1,i0)') '.', iproc
       ! <RN
       file_proc = trim(trim(file_proc)//adjustl(suffix))

       istatus = nf90_create (file_proc, NF90_CLOBBER, ncid)
       if (debug) write (error_unit(),*) &
            'created netcdf file ', trim(file_proc), ' with ncid: ', ncid, &
            ' in agk_save_for_restart'

       if (istatus /= NF90_NOERR) then
          ierr = error_unit()
          write(ierr,*) "nf90_create error: ", nf90_strerror(istatus)
          goto 1
       end if

       if (n_elements > 0) then
          istatus = nf90_def_dim (ncid, "theta", 2*ntgrid+1, thetaid)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_dim theta error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_dim (ncid, "sign", 2, signid)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_dim sign error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_dim (ncid, "glo", n_elements, gloid)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_dim glo error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_dim (ncid, "aky", naky, kyid)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_dim aky error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_dim (ncid, "akx", nakx, kxid)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_dim akx error: ", nf90_strerror(istatus)
             goto 1
          end if
       end if

       if (netcdf_real == 0) netcdf_real = get_netcdf_code_precision()

       istatus = nf90_def_var (ncid, "t0", netcdf_real, t0id)
       if (istatus /= NF90_NOERR) then
          ierr = error_unit()
          write(ierr,*) "nf90_def_var t0 error: ", nf90_strerror(istatus)
          goto 1
       end if

       istatus = nf90_def_var (ncid, "delt0", netcdf_real, delt0id)
       if (istatus /= NF90_NOERR) then
          ierr = error_unit()
          write(ierr,*) "nf90_def_var delt0 error: ", nf90_strerror(istatus)
          goto 1
       end if
    
       if (ant_on) then
          istatus = nf90_def_dim (ncid, "nk_stir", nk_stir, nk_stir_dim)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_dim nk_stir error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_var (ncid, "a_ant_r", netcdf_real, nk_stir_dim, a_antr_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var a_ant_r error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_var (ncid, "a_ant_i", netcdf_real, nk_stir_dim, a_anti_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var a_ant_i error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_var (ncid, "b_ant_r", netcdf_real, nk_stir_dim, b_antr_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var b_ant_r error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_var (ncid, "b_ant_i", netcdf_real, nk_stir_dim, b_anti_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var b_ant_i error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_dim (ncid, "nseed", nseed, nseed_dim)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_dim nseed error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_var (ncid, "init_seed", netcdf_real, nseed_dim, init_seed_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var init_seed error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_var (ncid, "fin_seed", netcdf_real, nseed_dim, fin_seed_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var fin_seed error: ", nf90_strerror(istatus)
             goto 1
          end if

       end if

       if (n_elements > 0) then
          istatus = nf90_def_var (ncid, "gr", netcdf_real, &
               (/ thetaid, signid, gloid /), gr_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var g error: ", nf90_strerror(istatus)
             goto 1
          end if

          istatus = nf90_def_var (ncid, "gi", netcdf_real, &
               (/ thetaid, signid, gloid /), gi_id)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write(ierr,*) "nf90_def_var g error: ", nf90_strerror(istatus)
             goto 1
          end if

          if (use_Phi) then
             istatus = nf90_def_var (ncid, "phi_r", netcdf_real, &
                  (/ thetaid, kxid, kyid /), phir_id)
             if (istatus /= NF90_NOERR) then
                ierr = error_unit()
                write(ierr,*) "nf90_def_var phi error: ", nf90_strerror(istatus)
                goto 1
             end if

             istatus = nf90_def_var (ncid, "phi_i", netcdf_real, &
                  (/ thetaid, kxid, kyid /), phii_id)
             if (istatus /= NF90_NOERR) then
                ierr = error_unit()
                write(ierr,*) "nf90_def_var phi error: ", nf90_strerror(istatus)
                goto 1
             end if
          end if

          if (use_Apar) then
             istatus = nf90_def_var (ncid, "apar_r", netcdf_real, &
                  (/ thetaid, kxid, kyid /), aparr_id)
             if (istatus /= NF90_NOERR) then
                ierr = error_unit()
                write(ierr,*) "nf90_def_var apar error: ", nf90_strerror(istatus)
                goto 1
             end if

             istatus = nf90_def_var (ncid, "apar_i", netcdf_real, &
                  (/ thetaid, kxid, kyid /), apari_id)
             if (istatus /= NF90_NOERR) then
                ierr = error_unit()
                write(ierr,*) "nf90_def_var apar error: ", nf90_strerror(istatus)
                goto 1
             end if
          end if

          if (use_Bpar) then
             istatus = nf90_def_var (ncid, "bpar_r", netcdf_real, &
                  (/ thetaid, kxid, kyid /), bparr_id)
             if (istatus /= NF90_NOERR) then
                ierr = error_unit()
                write(ierr,*) "nf90_def_var bparr error: ", nf90_strerror(istatus)
                goto 1
             end if

             istatus = nf90_def_var (ncid, "bpar_i", netcdf_real, &
                  (/ thetaid, kxid, kyid /), bpari_id)
             if (istatus /= NF90_NOERR) then
                ierr = error_unit()
                write(ierr,*) "nf90_def_var bpari error: ", nf90_strerror(istatus)
                goto 1
             end if
          end if
       end if

       istatus = nf90_enddef (ncid)
       if (istatus /= NF90_NOERR) then
          ierr = error_unit()
          write(ierr,*) "nf90_enddef error: ", nf90_strerror(istatus)
          goto 1
       end if
       if (debug) then
          ierr = error_unit()
          write (ierr,*) 'initialized became true in agk_save'
          write (ierr,*) 'ncid =', ncid
          write (ierr,*) 't0id =', t0id
       end if
    end if

    if (debug) write (error_unit(),*) 'before nf90_put_var (t0id)'
    istatus = nf90_put_var (ncid, t0id, t0)
    if (istatus /= NF90_NOERR) then
       ierr = error_unit()
       write (ierr,*) "nf90_put_var t0 error: ", nf90_strerror(istatus)
       write (ierr,*) 'ncid =', ncid
       write (ierr,*) 't0id =', t0id
       goto 1
    end if

    istatus = nf90_put_var (ncid, delt0id, delt0)
    if (istatus /= NF90_NOERR) then
       ierr = error_unit()
       write (ierr,*) "nf90_put_var delt0 error: ", nf90_strerror(istatus)
       goto 1
    end if

1   continue
    if (istatus /= NF90_NOERR) then
       i = nf90_close (ncid)
       if (debug) write (ierr,*) 'close file and return from save_for_restart'
       return
    end if

    if (n_elements > 0) then

       if (ant_on) then

          if (.not. allocated(atmp)) then
             allocate (atmp(nk_stir)); call alloc8(r1=atmp,v="atmp")
          end if
          atmp = real(a_ant)
          istatus = nf90_put_var (ncid, a_antr_id, atmp)

          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write (ierr,*) "nf90_put_var a_antr error: ", &
                  nf90_strerror(istatus), ' ', iproc
          end if

          atmp = aimag(a_ant)
          istatus = nf90_put_var (ncid, a_anti_id, atmp)

          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write (ierr,*) "nf90_put_var a_anti error: ", &
                  nf90_strerror(istatus), ' ', iproc
          end if

          atmp = real(b_ant)
          istatus = nf90_put_var (ncid, b_antr_id, atmp)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write (ierr,*) "nf90_put_var b_antr error: ", &
                  nf90_strerror(istatus), ' ', iproc
          end if

          atmp = aimag(b_ant)
          istatus = nf90_put_var (ncid, b_anti_id, atmp)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write (ierr,*) "nf90_put_var b_anti error: ", &
                  nf90_strerror(istatus), ' ', iproc
          end if
          !Write initial random number seed used for this run (integer vector)
          !NOTE: Convert from integer to real to save init_seed
          allocate(r_init_seed(1:nseed)); r_init_seed=0.
          r_init_seed(:)=real(init_seed(:))
          istatus = nf90_put_var (ncid, init_seed_id, r_init_seed)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write (ierr,*) "nf90_put_var init_seed error: ", &
                  nf90_strerror(istatus), ' ', iproc
          end if
          !Write final random number seed used for this run (integer vector)
          !NOTE: Convert from integer to real to save init_seed
          allocate(r_fin_seed(1:nseed)); r_fin_seed=0.
          r_fin_seed(:)=real(fin_seed(:))
          istatus = nf90_put_var (ncid, fin_seed_id, r_fin_seed)
          if (istatus /= NF90_NOERR) then
             ierr = error_unit()
             write (ierr,*) "nf90_put_var fin_seed error: ", &
                  nf90_strerror(istatus), ' ', iproc
          end if
!          deallocate (atmp)
          call dealloc8(r1=atmp,v="atmp")
          
          deallocate(r_init_seed,r_fin_seed)
       end if

       allocate (tmpr(2*ntgrid+1, 2, g_lo%llim_proc:g_lo%ulim_alloc)) ; call alloc8 (r3=tmpr, v="tmpr")
       
       tmpr = real(g)
       istatus = nf90_put_var (ncid, gr_id, tmpr)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, gr_id)
       
       tmpr = aimag(g)
       istatus = nf90_put_var (ncid, gi_id, tmpr)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, gi_id)
       
       call dealloc8 (r3=tmpr, v="tmpr") 

       allocate (ftmpr (2*ntgrid+1, nakx, naky)) ; call alloc8 (r3=ftmpr, v="ftmpr")
       allocate (ftmpi (2*ntgrid+1, nakx, naky)) ; call alloc8 (r3=ftmpi, v="ftmpi")

       if (use_Phi) then
          ftmpr = real(phinew)
          istatus = nf90_put_var (ncid, phir_id, ftmpr)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, phir_id)

          ftmpi = aimag(phinew)
          istatus = nf90_put_var (ncid, phii_id, ftmpi)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, phii_id)
       end if

       if (use_Apar) then
          ftmpr = real(aparnew)
          istatus = nf90_put_var (ncid, aparr_id, ftmpr)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, aparr_id)

          ftmpi = aimag(aparnew)
          istatus = nf90_put_var (ncid, apari_id, ftmpi)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, apari_id)
       end if

       if (use_Bpar) then
          ftmpr = real(bparnew)
          istatus = nf90_put_var (ncid, bparr_id, ftmpr)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, bparr_id)

          ftmpi = aimag(bparnew)
          istatus = nf90_put_var (ncid, bpari_id, ftmpi)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, bpari_id)
       end if

       call dealloc8 (r3=ftmpr, v="ftmpr") 
       call dealloc8 (r3=ftmpi, v="ftmpi")

    end if

    if (exit) then
       i = nf90_close (ncid)
    else
       i = nf90_sync (ncid)
       if (i /= NF90_NOERR) &
            call netcdf_error (istatus, message='nf90_sync error')
    end if

# else /* not defined NETCDF */

    if (proc0) write (error_unit(),*) &
         'WARNING: agk_save_for_restart is called without netcdf library'

# endif /* ifdef NETCDF */

  end subroutine agk_save_for_restart

  subroutine agk_save_for_restart_energy(array)

    use file_utils, only: error_unit
    use mp, only: proc0
    implicit none
    real, dimension(:), intent(in) :: array
# ifdef NETCDF
    integer :: istatus
    integer :: i, ierr

    istatus = nf90_create (restart_file_ene, NF90_CLOBBER, ncenid)

    if (debug) write (error_unit(),*) &
         'created netcdf file ', trim(restart_file_ene), ' with ncenid: ', &
         ncenid, ' in agk_save_for_restart_energy'

    if (istatus /= NF90_NOERR) then
       ierr = error_unit()
       write(ierr,*) "nf90_create error: ", nf90_strerror(istatus)
       goto 1
    end if

    istatus = nf90_def_dim (ncenid, "array0", size(array), arraydim)
    if (istatus /= NF90_NOERR) then
       ierr = error_unit()
       write(ierr,*) "nf90_def_dim array0 dim error: ", nf90_strerror(istatus)
       goto 1
    end if

    if (netcdf_real == 0) netcdf_real = get_netcdf_code_precision()

    istatus = nf90_def_var (ncenid, "array0", netcdf_real, arraydim, array_id)
    if (istatus /= NF90_NOERR) then
       ierr = error_unit()
       write(ierr,*) "nf90_def_var array0 error: ", nf90_strerror(istatus)
       goto 1
    end if

    istatus = nf90_enddef (ncenid)
    if (istatus /= NF90_NOERR) then
       ierr = error_unit()
       write(ierr,*) "nf90_enddef error: ", nf90_strerror(istatus)
       goto 1
    end if

1   continue
    if (istatus /= NF90_NOERR) then
       i = nf90_close (ncenid)
       if (debug) write (ierr,*) 'close file and return from save_for_restart_energy'
       return
    end if

    istatus = nf90_put_var (ncenid, array_id, array)
    if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncenid, array_id)
       
    i = nf90_close (ncenid)

# else /* not defined NETCDF */

    if (proc0) write (error_unit(),*) &
         'WARNING: agk_save_for_restart_energy is called without netcdf library'

# endif /* ifdef NETCDF */
    
  end subroutine agk_save_for_restart_energy

  subroutine agk_restore_many (g, scale, istatus, use_Phi, use_Apar, use_Bpar, many)
# ifdef NETCDF
    use mp, only: proc0, iproc, nproc
    use fields_arrays, only: phinew, aparnew, bparnew
    use fields_arrays, only: phi, apar, bpar
    use kgrids, only: naky, nakx
    use agk_mem, only: alloc8, dealloc8
# endif
    use theta_grid, only: ntgrid
    use agk_layouts, only: g_lo
    use file_utils, only: error_unit
    implicit none
    complex, dimension (-ntgrid:,:,g_lo%llim_proc:), intent (out) :: g
    real, intent (in) :: scale
# ifdef NETCDF
    integer, intent (out) :: istatus
# else
    integer :: istatus
# endif
    logical, intent (in) :: use_Phi, use_Apar, use_Bpar
    logical, intent (in) :: many
# ifdef NETCDF
    character (306) :: file_proc
    character (10) :: suffix
    character (10)  :: suffix_format
    integer :: n_elements
    integer :: i, ierr
    real :: fac

    n_elements = g_lo%ulim_proc-g_lo%llim_proc+1
    if (n_elements <= 0) return

    if (.not.initialized) then

       file_proc = trim(restart_file)

       call get_suffix_format(nproc,suffix_format)
       write (suffix,suffix_format) '.', iproc
!       write (suffix,'(a1,i0)') '.', iproc
       file_proc = trim(trim(file_proc)//adjustl(suffix))

    ! file open
       istatus = nf90_open (file_proc, NF90_NOWRITE, ncid)
       if (debug) write (error_unit(),*) &
            'opened netcdf file ', trim(file_proc), ' with ncid: ', ncid, &
            ' in agk_restore_many'
       if (istatus /= NF90_NOERR) &
          call netcdf_error (istatus, file=file_proc, abort=.true.)

       if (netcdf_real == 0) netcdf_real = get_netcdf_code_precision()
       call check_netcdf_file_precision (ncid)

       ! inquire dimension id
       istatus = nf90_inq_dimid (ncid, "theta", thetaid)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, dim='theta')

       istatus = nf90_inq_dimid (ncid, "sign", signid)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, dim='sign')

       istatus = nf90_inq_dimid (ncid, "glo", gloid)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, dim='glo')

       istatus = nf90_inq_dimid (ncid, "aky", kyid)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, dim='aky')

       istatus = nf90_inq_dimid (ncid, "akx", kxid)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, dim='akx')

       ! inquire dimension size
       istatus = nf90_inquire_dimension (ncid, thetaid, len=i)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, dimid=thetaid)
       if (i /= 2*ntgrid + 1) write(*,*) 'Restart error: ntgrid=? ',i,' : ',ntgrid,' : ',iproc

       istatus = nf90_inquire_dimension (ncid, signid, len=i)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, dimid=signid)
       if (i /= 2) write(*,*) 'Restart error: sign=? ',i,' : ',iproc

       istatus = nf90_inquire_dimension (ncid, gloid, len=i)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, dimid=gloid)
       if (i /= g_lo%ulim_proc-g_lo%llim_proc+1) write(*,*) 'Restart error: glo=? ',i,' : ',iproc

       istatus = nf90_inquire_dimension (ncid, kyid, len=i)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, dimid=kyid)
       if (i /= naky) write(*,*) 'Restart error: naky=? ',i,' : ',naky,' : ',iproc

       istatus = nf90_inquire_dimension (ncid, kxid, len=i)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, dimid=kxid)
       if (i /= nakx) write(*,*) 'Restart error: nakx=? ',i,' : ',nakx,' : ',iproc

       ! inquire variable id
       if (use_Phi) then
          istatus = nf90_inq_varid (ncid, "phi_r", phir_id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='phi_r')

          istatus = nf90_inq_varid (ncid, "phi_i", phii_id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='phi_i')
       end if

       if (use_Apar) then
          istatus = nf90_inq_varid (ncid, "apar_r", aparr_id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='apar_r')

          istatus = nf90_inq_varid (ncid, "apar_i", apari_id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='apar_i')
       end if

       if (use_Bpar) then
          istatus = nf90_inq_varid (ncid, "bpar_r", bparr_id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='bpar_r')

          istatus = nf90_inq_varid (ncid, "bpar_i", bpari_id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='bpar_i')
       end if

       istatus = nf90_inq_varid (ncid, "gr", gr_id)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='gr')

       istatus = nf90_inq_varid (ncid, "gi", gi_id)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='gi')

    end if  ! if (.not.initialized)

    ! read variables
    allocate (tmpr (2*ntgrid+1, 2, g_lo%llim_proc:g_lo%ulim_alloc)) ; call alloc8 (r3=tmpr, v="tmpr")
    allocate (tmpi (2*ntgrid+1, 2, g_lo%llim_proc:g_lo%ulim_alloc)) ; call alloc8 (r3=tmpi, v="tmpi")

    tmpr = 0.; tmpi = 0.
    istatus = nf90_get_var (ncid, gr_id, tmpr)
    if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, gr_id)

    istatus = nf90_get_var (ncid, gi_id, tmpi)
    if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, gi_id)

    g = cmplx(tmpr, tmpi)

    call dealloc8 (r3=tmpr, v="tmpr") 
    call dealloc8 (r3=tmpi, v="tmpi") 

    allocate (ftmpr (2*ntgrid+1, nakx, naky)) ; call alloc8 (r3=ftmpr, v="ftmpr")
    allocate (ftmpi (2*ntgrid+1, nakx, naky)) ; call alloc8 (r3=ftmpi, v="ftmpi")

    if (use_Phi) then
       istatus = nf90_get_var (ncid, phir_id, ftmpr)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, phir_id)

       istatus = nf90_get_var (ncid, phii_id, ftmpi)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, phii_id)

       phinew = cmplx(ftmpr, ftmpi)
    end if

    if (use_Apar) then
       istatus = nf90_get_var (ncid, aparr_id, ftmpr)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, aparr_id)
       
       istatus = nf90_get_var (ncid, apari_id, ftmpi)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, apari_id)

       aparnew = cmplx(ftmpr, ftmpi)
    end if

    if (use_Bpar) then
       istatus = nf90_get_var (ncid, bparr_id, ftmpr)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, bparr_id)
       
       istatus = nf90_get_var (ncid, bpari_id, ftmpi)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, bpari_id)
       
       bparnew = cmplx(ftmpr, ftmpi)
    end if

    call dealloc8 (r3=ftmpr, v="ftmpr") 
    call dealloc8 (r3=ftmpi, v="ftmpi") 
       
    if (scale > 0.) then
       g = g*scale
       phinew = phinew*scale
       aparnew = aparnew*scale
       bparnew = bparnew*scale
    else
       fac = - scale/(maxval(abs(phinew)))
       g = g*fac
       phinew = phinew*fac
       aparnew = aparnew*fac
       bparnew = bparnew*fac
    end if

    phi = phinew; apar = aparnew; bpar = bparnew

    if (.not.initialized) then
       istatus = nf90_close (ncid)
       if (istatus /= NF90_NOERR) then
          ierr = error_unit()
          write(ierr,*) "nf90_close error: ", nf90_strerror(istatus),' ',iproc
       end if
    end if

# else /* not defined NETCDF */

    write (error_unit(),*) &
         'ERROR: agk_restore_many is called without netcdf'

# endif /* ifdef NETCDF */

  end subroutine agk_restore_many


  subroutine agk_restore_energy (array)

    use file_utils, only: error_unit
    implicit none
    real, dimension(:), intent(out) :: array
# ifdef NETCDF
    integer :: istatus
    integer :: i, ierr


    istatus = nf90_open (restart_file_ene, NF90_NOWRITE, ncenid)
    if (debug) write (error_unit(),*) &
         'opened netcdf file ', trim(restart_file_ene), ' with ncenid: ', &
         ncenid, ' in agk_restore_energy'

    if (istatus /= NF90_NOERR) then
       !NOTE GGH 14 FEB 2012: Changed to abort=false to allow code to run when
       !       nc_ene file is missing
       call netcdf_error (istatus, file=restart_file_ene, abort=.false.)
       array(:)=0.
    else
       if (netcdf_real == 0) netcdf_real = get_netcdf_code_precision()
       
       ! inquire dimension id
       istatus = nf90_inq_dimid (ncenid, "array0", arraydim)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, dim='array0')
       
       ! inquire dimension size
       istatus = nf90_inquire_dimension (ncenid, arraydim, len=i)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncenid, dimid=arraydim) 
       if (i /= 2) write(*,*) 'Restart error: array size=? ',i
       
       ! inquire variable id
       istatus = nf90_inq_varid (ncenid, "array0", array_id)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='array0')
       
       istatus = nf90_get_var (ncenid, array_id, array)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncenid, array_id)
       
       istatus = nf90_close (ncenid)
       if (istatus /= NF90_NOERR) then
          ierr = error_unit()
          write(ierr,*) "nf90_close error: ", nf90_strerror(istatus)
       end if
    endif
# else /* not defined NETCDF */

    write (error_unit(),*) &
         'ERROR: agk_restore_energy is called without netcdf'

# endif /* ifdef NETCDF */

  end subroutine agk_restore_energy

  subroutine init_save (file,numdigits,sep_em_in,sep_em_out)
    character(300), intent (in) :: file
    integer, intent (in) :: numdigits         
    logical, intent (in) :: sep_em_in         
    logical, intent (in) :: sep_em_out       
    
    restart_file = file
    restart_file_ene = trim(file)//'_ene'
    ndigits=numdigits
    separate_em_in=sep_em_in
    separate_em_out=sep_em_out

  end subroutine init_save

  subroutine init_dt (delt0, istatus)

# ifdef NETCDF
    use mp, only: nproc, proc0, broadcast, iproc
    use file_utils, only: error_unit
# endif
    implicit none
    real, intent (in out) :: delt0
# ifndef NETCDF
    integer :: istatus
# else
    integer, intent (out) :: istatus
    character (10) :: suffix
    character (10)  :: suffix_format
    character (306) :: file_proc

    if (proc0) then

       if (.not. initialized) then

          file_proc = trim(restart_file)
          call get_suffix_format(nproc,suffix_format)
          write (suffix,suffix_format) '.', iproc
          file_proc = trim(trim(file_proc)//adjustl(suffix))

          istatus = nf90_open (file_proc, NF90_NOWRITE, ncid)
          if (debug) write (error_unit(),*) &
               'opened netcdf file ', trim(file_proc), ' with ncid: ', ncid
          if (istatus /= NF90_NOERR) call netcdf_error (istatus,file=file_proc)

          istatus = nf90_inq_varid (ncid, "delt0", delt0id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='delt0')

       end if

       istatus = nf90_get_var (ncid, delt0id, delt0)
       if (istatus /= NF90_NOERR) then
          call netcdf_error (istatus, ncid, delt0id, message=' in init_dt')
          delt0 = -1.
       endif

       if (.not.initialized) istatus = nf90_close (ncid)
    endif

    call broadcast (istatus)
    call broadcast (delt0)

# endif

  end subroutine init_dt

! This routine gets a_ant and b_ant for proc 0 only!!
  subroutine init_ant_amp (a_ant, b_ant, nk_stir, nseed, noseed, init_seed, istatus)
!
! TT: This subroutine may cause a problem with ncid because of the recent
! TT: change made in save_for_restart not closing file
!
# ifdef NETCDF
    use file_utils, only: error_unit
    use constants, only: zi
# endif
    use mp, only: proc0,nproc
    use agk_mem, only: alloc8, dealloc8
    implicit none
    complex, dimension(:), intent (in out) :: a_ant, b_ant
    integer, intent (in) :: nk_stir
    integer, intent(in) :: nseed
    logical, intent(out) :: noseed
    integer, dimension(:), intent(out) :: init_seed
    integer, intent (out) :: istatus
    real, allocatable, dimension(:) :: r_fin_seed
# ifdef NETCDF
    character (10) :: suffix
    character (10)  :: suffix_format
    character (306) :: file_proc
    integer :: ierr, i

    if (proc0) then
       a_ant = 0. ; b_ant = 0.

          file_proc = trim(restart_file)
          call get_suffix_format(nproc,suffix_format)
          write (suffix,suffix_format) '.', 0
          file_proc = trim(trim(file_proc)//adjustl(suffix))

!          file_proc=trim(trim(restart_file)//'.0')

       istatus = nf90_open (file_proc, NF90_NOWRITE, ncid)
       if (debug) write (error_unit(),*) &
            'opened netcdf file ', trim(file_proc), ' with ncid: ', ncid
       if (istatus /= NF90_NOERR) then
          ierr = error_unit()
          write(ierr,*) "nf90_open in init_ant_amp error: ", nf90_strerror(istatus) 
          write(ierr,*) "If you did not intend for this to be a restarted run with an external antenna,"
          write(ierr,*) "you may ignore the error message above."
          return
       endif

       istatus = nf90_inq_dimid (ncid, "nk_stir", nk_stir_dim)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, dim='nk_stir')

       istatus = nf90_inquire_dimension (ncid, nk_stir_dim, len=i)
       if (istatus /= NF90_NOERR) &
            call netcdf_error (istatus, ncid, dimid=nk_stir_dim)
       if (i /= nk_stir) write(error_unit(),*) 'Restart error: nk_stir=? ',i,' : ',nk_stir

       ! inquire variable id
       istatus = nf90_inq_varid (ncid, "a_ant_r", a_antr_id)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='a_ant_r')

       istatus = nf90_inq_varid (ncid, "a_ant_i", a_anti_id)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='a_ant_i')

       istatus = nf90_inq_varid (ncid, "b_ant_r", b_antr_id)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='b_ant_r')

       istatus = nf90_inq_varid (ncid, "b_ant_i", b_anti_id)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='b_ant_i')

       istatus = nf90_inq_dimid (ncid, "nseed", nseed_dim)
       if (istatus /= NF90_NOERR) then 
          noseed =.true.
          call netcdf_error (istatus, var='nseed')
       endif
       !If no seed in restart file, skip next statements
       if (.not. noseed) then
          istatus = nf90_inquire_dimension (ncid, nseed_dim, len=i)
          if (istatus /= NF90_NOERR) &
               call netcdf_error (istatus, ncid, dimid=nseed_dim)
          if (i /= nseed) then
             write(error_unit(),*) 'Restart error: nseed=? ',i,' : ',nseed
             write(error_unit(),*) 'Therefore, we choose to randomize seed'
             noseed=.true.
          endif

          istatus = nf90_inq_varid (ncid, "fin_seed", fin_seed_id)
          if (istatus /= NF90_NOERR) then 
             noseed =.true.
             call netcdf_error (istatus, var='fin_seed')
          endif
       endif

       if (.not. allocated(atmp)) then
          allocate (atmp(nk_stir)); call alloc8(r1=atmp,v="atmp")
       end if
       atmp = 0.

       ! read variables
       istatus = nf90_get_var (ncid, a_antr_id, atmp)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, a_antr_id)
       a_ant = atmp

       istatus = nf90_get_var (ncid, a_anti_id, atmp)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, a_anti_id)
       a_ant = a_ant + zi * atmp

       istatus = nf90_get_var (ncid, b_antr_id, atmp)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, b_antr_id)
       b_ant = atmp

       istatus = nf90_get_var (ncid, b_anti_id, atmp)
       if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, b_anti_id)
       b_ant = b_ant + zi * atmp

       if (.not. noseed) then
          !Read final seed from last run directly into init_seed
          !NOTE: Convert from saved real init_seed  to integer 
          allocate(r_fin_seed(1:nseed)); r_fin_seed=0.
          istatus = nf90_get_var (ncid, fin_seed_id, r_fin_seed)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, ncid, fin_seed_id)
          init_seed(:)=nint(r_fin_seed(:))
          deallocate(r_fin_seed)
       endif

!       deallocate (atmp)
       call dealloc8(r1=atmp,v="atmp")
       istatus = nf90_close (ncid)       
    endif

# else

    if (proc0) istatus = 2

# endif

  end subroutine init_ant_amp

  subroutine init_tstart (tstart, istatus)

# ifdef NETCDF
    use mp, only: nproc, proc0, broadcast
    use file_utils, only: error_unit
# endif
    implicit none
    real, intent (in out) :: tstart
# ifndef NETCDF
    integer :: istatus
# else
    integer, intent (out) :: istatus
    character (10) :: suffix
    character (10)  :: suffix_format
    character (306) :: file_proc

    if (proc0) then

       if (.not.initialized) then

          file_proc = trim(restart_file)
          call get_suffix_format(nproc,suffix_format)
          write (suffix,suffix_format) '.', 0
          file_proc = trim(trim(file_proc)//adjustl(suffix))

!          file_proc=trim(trim(restart_file)//'.0')

          istatus = nf90_open (file_proc, NF90_NOWRITE, ncid)
          if (debug) write (error_unit(),*) &
               'opened netcdf file ', trim(file_proc), ' with ncid: ', ncid
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, file=file_proc)

          istatus = nf90_inq_varid (ncid, "t0", t0id)
          if (istatus /= NF90_NOERR) call netcdf_error (istatus, var='t0')

       end if

       istatus = nf90_get_var (ncid, t0id, tstart)
       if (istatus /= NF90_NOERR) then
          call netcdf_error (istatus, ncid, t0id, message=' in init_tstart')
          tstart = -1.
       end if

       if (.not.initialized) then
          istatus = nf90_close (ncid)
          if (debug) print *, 'ncid closed in init_tstart'
       end if
    endif

    call broadcast (istatus)
    call broadcast (tstart)

# endif

  end subroutine init_tstart

!-----------------------------------------------------------------------------
  subroutine get_suffix_format(nproc,suffix_format)
    ! Determine the optimal number of digits in netcdf restart files
    implicit none
    !Passed
    integer, intent(in) :: nproc !Number of processors for this job
    character (10), intent(out)  :: suffix_format
    !Local
    integer :: numdig            !Number of digits to use in restart filename

    !For NetCDF restart files, use the minimum number of digits to handle nproc
    !    unless ndigits has been set in init_g_knobs namelist (default is -1)
    if (nproc==1) then
       numdig=1
    else
       if (ndigits .eq. -1) then
          numdig=int(alog10(real(nproc-1)))+1
       else
          numdig=ndigits
       endif
    end if

    write(suffix_format,'(a,i1,a,i1,a)')'(a1,i',numdig,'.',numdig,')'
  end subroutine get_suffix_format

end module agk_save
